from celery.schedules import crontab, timedelta

class ScheduleCron:
    def __init__(self):
        self.schedule_cron = None

    def cron(self, cron_schedule):
        self.schedule_cron = crontab(cron_schedule)
        return self

    def everySecond(self):
        self.schedule_cron = timedelta(seconds=1)
        return self

    def everyTwoSeconds(self):
        self.schedule_cron = timedelta(seconds=2)
        return self

    def everyFiveSeconds(self):
        self.schedule_cron = timedelta(seconds=5)
        return self

    def everyTenSeconds(self):
        self.schedule_cron = timedelta(seconds=10)
        return self

    def everyFifteenSeconds(self):
        self.schedule_cron = timedelta(seconds=15)
        return self

    def everyTwentySeconds(self):
        self.schedule_cron = timedelta(seconds=20)
        return self

    def everyThirtySeconds(self):
        self.schedule_cron = timedelta(seconds=30)
        return self

    def everyMinute(self):
        self.schedule_cron = crontab(minute='*')
        return self

    def everyTwoMinutes(self):
        self.schedule_cron = crontab(minute='*/2')
        return self

    def everyThreeMinutes(self):
        self.schedule_cron = crontab(minute='*/3')
        return self

    def everyFourMinutes(self):
        self.schedule_cron = crontab(minute='*/4')
        return self

    def everyFiveMinutes(self):
        self.schedule_cron = crontab(minute='*/5')
        return self

    def everyTenMinutes(self):
        self.schedule_cron = crontab(minute='*/10')
        return self

    def everyFifteenMinutes(self):
        self.schedule_cron = crontab(minute='*/15')
        return self

    def everyThirtyMinutes(self):
        self.schedule_cron = crontab(minute='*/30')
        return self

    def hourly(self):
        self.schedule_cron = crontab(minute=0, hour='*')
        return self

    def hourlyAt(self, minute):
        self.schedule_cron = crontab(minute=minute, hour='*')
        return self

    def everyOddHour(self, minutes=0):
        self.schedule_cron = crontab(minute=minutes, hour='1-23/2')
        return self

    def everyTwoHours(self, minutes=0):
        self.schedule_cron = crontab(minute=minutes, hour='*/2')
        return self

    def everyThreeHours(self, minutes=0):
        self.schedule_cron = crontab(minute=minutes, hour='*/3')
        return self

    def everyFourHours(self, minutes=0):
        self.schedule_cron = crontab(minute=minutes, hour='*/4')
        return self

    def everySixHours(self, minutes=0):
        self.schedule_cron = crontab(minute=minutes, hour='*/6')
        return self

    def daily(self):
        self.schedule_cron = crontab(minute=0, hour=0)
        return self

    def dailyAt(self, time):
        hour, minute = map(int, time.split(':'))
        self.schedule_cron = crontab(minute=minute, hour=hour)
        return self

    def twiceDaily(self, first, second):
        self.schedule_cron = crontab(minute=0, hour=(first, second))
        return self

    def twiceDailyAt(self, first, second, minute):
        self.schedule_cron = crontab(minute=minute, hour=(first, second))
        return self

    def weekly(self):
        self.schedule_cron = crontab(minute=0, hour=0, day_of_week=0)
        return self

    def weeklyOn(self, day, time):
        hour, minute = map(int, time.split(':'))
        self.schedule_cron = crontab(minute=minute, hour=hour, day_of_week=day)
        return self

    def monthly(self):
        self.schedule_cron = crontab(minute=0, hour=0, day_of_month=1)
        return self

    def monthlyOn(self, day, time):
        hour, minute = map(int, time.split(':'))
        self.schedule_cron = crontab(minute=minute, hour=hour, day_of_month=day)
        return self

    def twiceMonthly(self, first, second, time):
        hour, minute = map(int, time.split(':'))
        self.schedule_cron = crontab(minute=minute, hour=hour, day_of_month=(first, second))
        return self

    def lastDayOfMonth(self, time):
        hour, minute = map(int, time.split(':'))
        self.schedule_cron = crontab(minute=minute, hour=hour, day_of_month=-1)
        return self

    def quarterly(self):
        self.schedule_cron = crontab(minute=0, hour=0, month_of_year=(1, 4, 7, 10), day_of_month=1)
        return self

    def quarterlyOn(self, month, time):
        hour, minute = map(int, time.split(':'))
        self.schedule_cron = crontab(minute=minute, hour=hour, month_of_year=month, day_of_month=1)
        return self

    def yearly(self):
        self.schedule_cron = crontab(minute=0, hour=0, day_of_month=1, month_of_year=1)
        return self

    def yearlyOn(self, month, day, time):
        hour, minute = map(int, time.split(':'))
        self.schedule_cron = crontab(minute=minute, hour=hour, day_of_month=day, month_of_year=month)
        return self


from fastapi import Request
from urllib.parse import urlparse

def paginate(request: Request, query, serilizer, the_page: int = 1, the_per_page: int = 10, wrap: str = "data", additional_data: dict = None):
    """Paginate the query.
    
    Args:
        request (Request): The FastAPI request object.
        query: The database query to paginate.
        serilizer: The serializer class to convert database objects.
        the_page (int, optional): Default page number. Defaults to 1.
        the_per_page (int, optional): Default items per page. Defaults to 10.
        wrap (str, optional): Key name for the data in response. Defaults to 'data'.
        additional_data (dict, optional): Additional data to include in the response. Defaults to None.
    """

    page = int(request.query_params.get("page", the_page))
    per_page = int(request.query_params.get("per_page", the_per_page))

    total = query.count()
    last_page = (total + per_page - 1) // per_page
    offset = (page - 1) * per_page
    paginated_query = query.offset(offset).limit(per_page).all()

    data = [serilizer.model_validate(item) for item in paginated_query]

    base_url = str(request.base_url)

    full_path = str(request.url)
    parsed_url = urlparse(full_path)
    path_without_query = f"{parsed_url.scheme}://{parsed_url.netloc}{parsed_url.path}"

    first_page_url = f"{path_without_query}?page=1&per_page={per_page}"
    last_page_url = f"{path_without_query}?page={last_page}&per_page={per_page}"
    next_page_url = f"{path_without_query}?page={page + 1}&per_page={per_page}" if page < last_page else None
    prev_page_url = f"{path_without_query}?page={page - 1}&per_page={per_page}" if page > 1 else None

    output = {
        "total": total,
        "per_page": per_page,
        "current_page": page,
        "last_page": last_page,
        "first_page_url": first_page_url,
        "last_page_url": last_page_url,
        "next_page_url": next_page_url,
        "prev_page_url": prev_page_url,
        "path": base_url,
        "from": offset + 1 if data else None,
        "to": offset + len(data) if data else None,
        wrap: data
    }

    if additional_data:
        if callable(additional_data):
            output["additional_data"] = additional_data(data)
        else:
            output["additional_data"] = additional_data

    return output



import bcrypt

def generate_password_hash(password: str) -> bytes:
    """
    Generate a bcrypt hash for the given password.
    
    Args:
        password (str): The plain text password to hash
        
    Returns:
        bytes: The hashed password
    """
    if not isinstance(password, str):
        raise TypeError("Password must be a string")
        
    if not password:
        raise ValueError("Password cannot be empty")
    
    # Convert the password to bytes and generate salt
    password_bytes = password.encode('utf-8')
    # Use bcrypt.gensalt(12) to specify rounds (default is 12)
    salt = bcrypt.gensalt(12)
    
    # Generate the hashed password
    hashed = bcrypt.hashpw(password_bytes, salt)

    hashed_decoded = hashed.decode('utf-8')
    
    return hashed_decoded

def compare_hashed_password(password: str, hashed_password: str | bytes) -> bool:
    """
    Compare a plain text password against a hashed password.
    
    Args:
        password (str): The plain text password to check
        hashed_password (str|bytes): The hashed password to compare against
        
    Returns:
        bool: True if passwords match, False otherwise
    """
    if not isinstance(password, str):
        raise TypeError("Password must be a string")
    
    try:
        # Convert password to bytes
        password_bytes = password.encode('utf-8')
        
        # Ensure hashed_password is bytes
        if isinstance(hashed_password, str):
            hashed_bytes = hashed_password.encode('utf-8')
        else:
            hashed_bytes = hashed_password
            
        # Verify the hash format
        if not hashed_bytes.startswith(b'$2b$') and not hashed_bytes.startswith(b'$2a$'):
            raise ValueError("Invalid hash format")
            
        return bcrypt.checkpw(password_bytes, hashed_bytes)
        
    except (ValueError, TypeError) as e:
        print(f"Password comparison failed: {str(e)}")
        return False


from fastapi import Request, BackgroundTasks
from pydantic import BaseModel, ValidationError
from functools import wraps
from fastapi_pundra.rest.helpers import the_query
from fastapi_pundra.rest.exceptions import ValidationException, BadRequestException

def dto(schema: BaseModel):
    def decorator(func):
        @wraps(func)
        async def wrapper(request: Request, background_tasks: BackgroundTasks = None, *args, **kwargs):
            try:
                request_data = await the_query(request)
                validated_data = schema(**request_data)
                request.state.validated_data = validated_data
                if background_tasks is not None:
                    return await func(request, background_tasks, *args, **kwargs)
                return await func(request,*args, **kwargs)
            except ValidationError as e:
                errors = {}
                for error in e.errors():
                    field = error["loc"][0]
                    message = field + " " + error["msg"]
                    if field not in errors:
                        errors[field] = []
                    errors[field].append(message)
                raise ValidationException(errors=errors)
            except ValueError:
                raise BadRequestException(message="Invalid JSON")

        return wrapper
    return decorator

__all__ = ["dto"]



import inspect
from .schedule_cron import ScheduleCron

class ScheduleConverter:
    def __init__(self, schedules):
        self.schedules = schedules
        self.result = {}

    def convert(self):
        for schedule in self.schedules:
            task_method =  schedule.task_data.get('task')
            name = schedule.name_data
            last_part = schedule.schedule_cron
            task_method_name = task_method.__name__
            module_name = inspect.getmodule(task_method).__name__
            
            self.result[name] = {
                'task': f'{module_name}.{task_method_name}',
                'schedule': last_part,
            }

        return self.result

class Schedule(ScheduleCron):
    def __init__(self):
        self.task_data = {}
        self.name_data = None
        self.schedule_cron = None

    def task(self, task):
        self.task_data['task'] = task
        return self

    def name(self, name):
        self.name_data = name
        return self


def bind_beat_schedule(schedules):
    data = ScheduleConverter(schedules).convert()
    return data


def paginate(query, page: int = 1, per_page: int = 10, additional_data: dict = None):
    total = query.count()
    last_page = (total + per_page - 1) // per_page
    
    # if page is greater than last page, set page to last page
    if page > last_page:
        page = last_page

    offset = (page - 1) * per_page
    paginated_query = query.offset(offset).limit(per_page).all()

    data = paginated_query

    next_page = page + 1 if page < last_page else None
    prev_page = page - 1 if page > 1 else None

    output = {
        "data": data,
        "pagination": {
            "total": total,
            "current_page": page,
            "next_page": next_page,
            "prev_page": prev_page,
            "per_page": per_page,
            "last_page": last_page,
            "from_item": offset + 1 if data else None,
            "to_item": offset + len(data) if data else None
        }
    }

    if additional_data:
        if callable(additional_data):
            output['additional_data'] = additional_data(data)
        else:
            output['additional_data'] = additional_data

    return output
    


from fastapi import BackgroundTasks
from fastapi_mail import FastMail, MessageSchema, ConnectionConfig
from dotenv import load_dotenv
from .mail_templating import EmailTemplates
from fastapi_pundra.common.helpers import base_path
import os
from typing import List

# Load environment variables from .env file
load_dotenv()


# Configure email connection
conf = ConnectionConfig(
    MAIL_USERNAME=os.getenv('MAIL_USERNAME', 'default_username'),
    MAIL_PASSWORD=os.getenv('MAIL_PASSWORD', 'default_password'),
    MAIL_FROM=os.getenv('MAIL_FROM_ADDRESS', 'noreply@example.com'),  # Default sender address
    MAIL_PORT=int(os.getenv('MAIL_PORT', '587')),  # Default to common SMTP port
    MAIL_SERVER=os.getenv('MAIL_HOST', 'smtp.gmail.com'),  # Default to Gmail SMTP server
    MAIL_STARTTLS=False,  # or False based on your configuration
    MAIL_SSL_TLS=False,   # or False based on your configuration
    USE_CREDENTIALS=True
)

def render_mail_template(template_name: str, context: dict = {}):
    # Initialize EmailTemplates with the correct directory
    project_base_path = os.getenv('PROJECT_BASE_PATH', 'app')

    template_dir = os.path.join(base_path(), project_base_path, 'templates', 'mails')
    templates = EmailTemplates(directory=template_dir)
    # Render the specified template with the given context
    return templates.render_template(template_name, context)

async def send_mail_util(subject: str, to: List[str], template_name: str, context: dict = None, cc: List[str] | str = None, bcc: List[str] | str = None, reply_to: List[str] | str = None):
    try:
        if not isinstance(to, list):
            to = [to]
        
        # Convert cc, bcc, reply_to to lists if they are strings
        if cc and not isinstance(cc, list):
            cc = [cc]
        if bcc and not isinstance(bcc, list):
            bcc = [bcc]
        if reply_to and not isinstance(reply_to, list):
            reply_to = [reply_to]
            
        message = MessageSchema(
            subject=subject,
            recipients=to,
            cc=cc,
            bcc=bcc,
            reply_to=reply_to,
            body=render_mail_template(template_name=template_name, context=context),
            subtype="html"
        )
        
        fm = FastMail(conf)
        await fm.send_message(message)
        return "Email has been sent"
    except Exception as e:
        print(f"Failed to send email: {str(e)}")
        return f'An error occurred while sending the email: {str(e)}'

async def send_mail_background(background_tasks: BackgroundTasks, subject: str, to: List[str], template_name: str, context: dict = None, cc: List[str] | str = None, bcc: List[str] | str = None, reply_to: List[str] | str = None):
    background_tasks.add_task(send_mail_util, subject, to, template_name, context, cc, bcc, reply_to)
    return "Email scheduled for sending"

async def send_mail(subject: str, to: List[str], template_name: str, context: dict = None, cc: List[str] | str = None, bcc: List[str] | str = None, reply_to: List[str] | str = None):
    return await send_mail_util(subject, to, template_name, context, cc, bcc, reply_to)




import os
from fastapi_pundra.common.helpers import base_path
from dotenv import load_dotenv
from sqlalchemy import text, Result
from collections import OrderedDict
from fastapi import Request
from urllib.parse import urlparse
from uuid import UUID
from datetime import datetime
from sqlalchemy.engine.result import Result

load_dotenv()


def raw_sql_fetch_all(query: Result) -> list[dict]:
    """Fetch all data from the query."""
    data = query.fetchall()
    result_list = []
    for row in data:
        row_dict = dict(row._asdict())
        # Convert UUID and datetime objects to strings
        for key, value in row_dict.items():
            if isinstance(value, UUID):
                row_dict[key] = str(value)
            elif isinstance(value, datetime):
                row_dict[key] = value.isoformat()
        result_list.append(row_dict)

    return result_list

def load_sql_file(file_path: str, sql_vars: dict = None) -> str:
    project_base_path = os.getenv('PROJECT_BASE_PATH', 'app')
    sql_files_path = 'sql_files'
    
    # Split the file path into components and join with SQL extension
    components = file_path.split(".")
    relative_path = os.path.join(*components) + ".sql"
    
    # Construct absolute path using base_path and project settings
    absolute_path = os.path.join(base_path(), project_base_path, sql_files_path, relative_path)
    
    with open(absolute_path, "r") as file:
        sql_content = file.read()
        
    # Replace SQL variables if provided
    if sql_vars:
        for key, value in sql_vars.items():
            placeholder = f"--sql_var:{key}"
            # Handle different types of values
            if isinstance(value, str):
                formatted_value = f"'{value}'"
            elif value is None:
                formatted_value = 'NULL'
            else:
                formatted_value = str(value)
            sql_content = sql_content.replace(placeholder, formatted_value)
    
    return text(sql_content)
    
def raw_sql_rest_paginate(request: Request, query_data: list, serializer = None, the_page: int = 1, the_per_page: int = 10, wrap: str = "data", additional_data: dict = None) -> dict:
    """Paginate raw SQL query results."""
    results = query_data

    page = int(request.query_params.get("page", the_page))
    per_page = int(request.query_params.get("per_page", the_per_page))

    # Calculate indices and total pages
    total = len(results)
    total_pages = (total + per_page - 1) // per_page
    start_index = (page - 1) * per_page
    end_index = page * per_page

    # Slice the results to get the paginated subset
    paginated_results = results[start_index:end_index]

    if serializer:
        paginated_results = [serializer(**item).model_dump() for item in paginated_results]

    # Build proper URLs using urlparse
    full_path = str(request.url)
    parsed_url = urlparse(full_path)
    path_without_query = f"{parsed_url.scheme}://{parsed_url.netloc}{parsed_url.path}"
    base_url = str(request.base_url)

    # Generate all pagination URLs
    first_page_url = f"{path_without_query}?page=1&per_page={per_page}"
    last_page_url = f"{path_without_query}?page={total_pages}&per_page={per_page}"
    next_page_url = f"{path_without_query}?page={page + 1}&per_page={per_page}" if page < total_pages else None
    prev_page_url = f"{path_without_query}?page={page - 1}&per_page={per_page}" if page > 1 else None

    output = OrderedDict([
        (wrap, paginated_results),
        ("total", total),
        ("per_page", per_page),
        ("current_page", page),
        ("last_page", total_pages),
        ("first_page_url", first_page_url),
        ("last_page_url", last_page_url),
        ("next_page_url", next_page_url),
        ("prev_page_url", prev_page_url),
        ("path", base_url),
        ("from", start_index + 1 if paginated_results else None),
        ("to", start_index + len(paginated_results) if paginated_results else None),
    ])

    if additional_data:
        if callable(additional_data):
            # Convert list items to a dictionary before passing to additional_data
            output["additional_data"] = additional_data(paginated_results)
        else:
            output["additional_data"] = additional_data

    return output

def raw_sql_fetch_one(query: Result, serializer = None) -> dict:
    """Fetch one data from the query."""
    data = query.fetchone()
    if data is None:
        return None
    # Convert Row to dictionary first
    data_dict = dict(data._asdict())
    if serializer:
        return serializer(**data_dict).model_dump()
    return data_dict

def raw_sql_paginate_gql(query_data: list, serializer = None, the_page: int = 1, the_per_page: int = 10, wrap: str = "data", additional_data: dict = None) -> dict:
    """Paginate raw SQL query results."""
    results = query_data
    
    # Calculate pagination values
    total = len(results)
    last_page = (total + the_per_page - 1) // the_per_page
    
    # Adjust page if it exceeds last_page
    page = min(the_page, last_page) if last_page > 0 else 1
    
    # Calculate indices
    start_index = (page - 1) * the_per_page
    end_index = page * the_per_page
    
    # Slice the results
    paginated_results = results[start_index:end_index]
    
    # Apply serializer if provided
    if serializer:
        paginated_results = [serializer(**item).model_dump() for item in paginated_results]
    
    # Calculate next and previous pages
    next_page = page + 1 if page < last_page else None
    prev_page = page - 1 if page > 1 else None
    
    output = {
        wrap: paginated_results,
        "pagination": {
            "total": total,
            "current_page": page,
            "next_page": next_page,
            "prev_page": prev_page,
            "per_page": the_per_page,
            "last_page": last_page,
            "from_item": start_index + 1 if paginated_results else None,
            "to_item": start_index + len(paginated_results) if paginated_results else None
        }
    }
    
    if additional_data:
        if callable(additional_data):
            output["additional_data"] = additional_data(paginated_results)
        else:
            output["additional_data"] = additional_data
    
    return output


import jinja2
import warnings
from os import PathLike
from .inline_css import inline_css

try:
    import jinja2
except ModuleNotFoundError:
    jinja2 = None


class EmailTemplates:
    def __init__(
        self,
        directory,
        *,
        context_processors=None,
        env=None,
        **env_options
    ):
        assert jinja2 is not None, "jinja2 must be installed to use EmailTemplates"
        assert directory or env, "either 'directory' or 'env' arguments must be passed"
        self.context_processors = context_processors or []
        if directory is not None:
            self.env = self._create_env(directory, **env_options)
        elif env is not None:
            self.env = env

        self._setup_env_defaults(self.env)

    def _create_env(
        self,
        directory,
        **env_options
    ):
        loader = jinja2.FileSystemLoader(directory)
        env_options.setdefault("loader", loader)
        env_options.setdefault("autoescape", True)

        return jinja2.Environment(**env_options)

    def _setup_env_defaults(self, env):
        pass

    def get_template(self, name):
        return self.env.get_template(name)

    def render_template(
        self,
        name,
        context=None,
    ):
        context = context or {}
        for context_processor in self.context_processors:
            context.update(context_processor())

        template = self.get_template(name)
        a =  template.render(**context)
        return inline_css(a)


from celery import shared_task
from typing import List
import asyncio
from .mail import send_mail

@shared_task(name="pundra_send_email_queue_task")
def send_email_queue_task(subject: str, to: List[str], template_name: str, context: dict, cc: List[str] | str = None, bcc: List[str] | str = None, reply_to: List[str] | str = None):
    # Create and run the event loop
    loop = asyncio.new_event_loop()
    asyncio.set_event_loop(loop)
    try:
        result = loop.run_until_complete(send_mail(
            subject=subject,
            to=to,
            template_name=template_name,
            context=context,
            cc=cc,
            bcc=bcc,
            reply_to=reply_to
        ))
        print(str(result))
        print(f"Sending email to {to} with subject {subject} and template {template_name} and context {context}")
        return "Email sent"
    finally:
        loop.close()